import ee, pandas as pd
from math import sqrt


def get_all_bands(image, date, polygon):
    '''
    Calculate all the Sentinel-2 bands mean values for a specific date and polygon in a Sentinel-2 image collection.
    
    Spectral bands refer to specific ranges of electromagnetic radiation (EMR) that are used in remote sensing 
    applications to capture information about the earth's surface. Spectral bands are usually defined by their
    wavelength or frequency, and are typically categorized into broad groups based on their spectral characteristics, 
    such as visible, near-infrared, and shortwave infrared.
    Each spectral band provides unique information about the reflectance and absorption properties of features on the
    earth's surface. For example, visible bands are sensitive to the reflectance of green vegetation, water, and soil,
    while near-infrared bands are sensitive to the reflectance of healthy vegetation. Finally, shortwave infrared bands
    can detect differences in moisture content and mineralogy.

    Args:
        image (ee.ImageCollection): The Sentinel 2 image collection filtered by date and bounds.
        date (pd.Timestamp): The acquisition date in Pandas Timestamp format.
        polygon (ee.Geometry): The field polygon geometry in Earth Engine format.
    
    Returns:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    '''
    # Filter image collection to get the image for the date
    image = ee.Image(image.filterDate(date.strftime('%Y-%m-%d'), (date + pd.Timedelta(days=1)).strftime('%Y-%m-%d')).first())

    # Mask out clouds and shadows
    image = image.updateMask(image.select('QA60').bitwiseAnd(2).neq(2))
    
    # Returns the list of mean bands values
    bands_list = ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8', 'B8A', 'B9', 'B11', 'B12']
    bands_means = {}
    for band in bands_list:
        bands_means[band] = image.reduceRegion(reducer=ee.Reducer.mean(), geometry=polygon).getInfo()[band]
    return bands_means


def calculate_vegetation_index(bands_means, type, id=None):
    '''
    Calculates the specified VI index for a given image, date, and polygon.
    
    While all of the listed VI indices are used to assess vegetation health and density, they differ in 
    the specific wavelengths of light that are used to calculate the index.

    Each of these indices may have different strengths and weaknesses depending on the specific vegetation types
    and environmental conditions being studied. For example, some indices may be more sensitive to vegetation 
    stress or changes in leaf structure than others. Therefore, researchers and practitioners may choose 
    different VI indices depending on their specific research questions and the available data.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        type (str): The type of vegetation index to calculate.
        id (int): The id of the specified type of vegetation index to calculate.
            
    Returns:
        float: the calculated VI value, for the specified date and polygon.
    '''
    # Return NDVI
    if (type == 'ND'):
        return (bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + bands_means['B4'])
    elif (type == 'NSND'):
        return (bands_means['B11'] - bands_means['B7']) / (bands_means['B11'] + bands_means['B7'])
    elif (type == 'GND'):
        return (bands_means['B8'] - bands_means['B3']) / (bands_means['B8'] + bands_means['B3'])
    elif (type == 'REND'):
        if (id == 1):
            return (bands_means['B5'] - bands_means['B4']) / (bands_means['B5'] + bands_means['B4'])
        elif (id == 2):
            return (bands_means['B6'] - bands_means['B4']) / (bands_means['B6'] + bands_means['B4'])
        elif (id == 3):
            return (bands_means['B7'] - bands_means['B4']) / (bands_means['B7'] + bands_means['B4'])
    elif (type == 'GRND'):
        return (bands_means['B8'] - (bands_means['B3'] + bands_means['B4'])) / (bands_means['B8'] + (bands_means['B3'] + bands_means['B4']))
    elif (type == 'GBND'):
        return (bands_means['B8'] - (bands_means['B3'] + bands_means['B2'])) / (bands_means['B8'] + (bands_means['B3'] + bands_means['B2']))


    # Return SAVI
    elif (type == 'SA'):
        L = 0.428
        return (bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + bands_means['B4'] + L) * (1 + L)
    elif (type == 'MSA'):
        return (2.0 * bands_means['B8'] + 1.0 - sqrt(((2.0 * bands_means['B8'] + 1.0) ** 2.0) - 8.0 * (bands_means['B8'] - bands_means['B4']))) / 2.0
    elif (type == 'OSA'):
        return (1.0 + 0.16) * (bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + bands_means['B4'] + 0.16)
    elif (type == 'TSA'):
        X, A, B = 0.114, 0.824, 0.421
        return (B * (bands_means['B8'] - B * bands_means['B4'] - A)) / (bands_means['B4'] + B * (bands_means['B8'] - A) + X * (1.0 + (B ** 2.0)))
    elif (type == 'ATSA'):
        return 1.22 * (bands_means['B8'] - 1.22 * bands_means['B4'] - 0.03) / (1.22 * bands_means['B8'] + bands_means['B4'] - 1.22 * 0.03 + 0.08 * (1.0 + (1.22 ** 2.0)))


    # Other VIs
    elif (type == 'A'):
        return (bands_means['B8'] * (1 - bands_means['B4'])*(bands_means['B8'] - bands_means['B4'])) ** 1/3
    elif (type == 'AR'):
        if (id == 1):
            return (bands_means['B8A'] - bands_means['B4'] - 0.069 * (bands_means['B4'] - bands_means['B2'])) / (bands_means['B8A'] + bands_means['B4'] - 0.069 * (bands_means['B4'] - bands_means['B2']))
        elif (id == 2):
            return -0.18 + 1.17 * ((bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + bands_means['B4']))
    elif (type == 'C'):
        return (bands_means['B8'] * bands_means['B4']) / (bands_means['B3'] ** 2.0)
    elif (type == 'CT'):
        return (((bands_means['B4'] - bands_means['B3']) / (bands_means['B4'] + bands_means['B3'])) + 0.5) / (abs(((bands_means['B4'] - bands_means['B3']) / (bands_means['B4'] + bands_means['B3']))) + 0.5 * sqrt(abs((((bands_means['B4'] - bands_means['B3']) / (bands_means['B4'] + bands_means['B3']))) + 0.5)))
    elif (type == 'D'):
        return bands_means['B8'] - bands_means['B4']
    elif (type == 'E'):
        if (id == 1):
            return 2.5 * (bands_means['B8'] - bands_means['B4']) / ((bands_means['B8'] + 6.0 * bands_means['B4'] - 7.5 * bands_means['B2']) + 1.0)
        elif (id == 2):
            return 2.4 * (bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + bands_means['B4'] + 1.0)
        elif (id == 3):
            return 2.5 * (bands_means['B8'] - bands_means['B4']) / (bands_means['B8'] + 2.4 * bands_means['B4'] + 1.0)
    elif (type == 'MT'):
        if (id == 1):
            return 1.2 * (1.2 * (bands_means['B8'] - bands_means['B3']) - 2.5 * (bands_means['B4'] - bands_means['B3']))
        elif (id == 2):
            return 1.5 * (1.2 * (bands_means['B8'] - bands_means['B3']) - 2.5 * (bands_means['B4'] - bands_means['B3'])) / sqrt(((2.0 * bands_means['B8'] + 1.0) ** 2.0) - (6.0 * bands_means['B8'] - 5.0 * sqrt(bands_means['B4'])) - 0.5)
    elif (type == 'R'):
        return bands_means['B8'] / bands_means['B4'] 
    elif (type == 'WDR'):
        return (0.1 * bands_means['B8'] - bands_means['B4']) / (0.1 * bands_means['B8'] + bands_means['B4'])


def calculate_exogenous_organic_matter_index(bands_means, id):
    '''
    Calculates the EOMIx (Exogenous Organic Matter Index) index for a given image, date, and polygon.

    The Exogenous Organic Matter Index (EOMI) is a soil quality indicator used to assess the content and quality of
    organic matter in soil. The EOMI is sensitive to the quantity and quality of exogenous organic matter, such as 
    crop residues, manure, and other organic inputs, as well as the presence of mineral and inorganic components in soil.
    For further details read the following paper: https://www.mdpi.com/2072-4292/13/9/1616.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the EOMI to be calculated (from 1 to 4).
        
    Returns:
        float: the calculated EOMIx value, for the specified date and polygon.
    '''
    # Calculate EOMI (different with respect to the passed id)
    if (id == 1):
        return (bands_means['B11'] - bands_means['B8A']) / (bands_means['B11'] + bands_means['B8A'])
    if (id == 2):
        return (bands_means['B12'] - bands_means['B4']) / (bands_means['B12'] + bands_means['B4'])
    if (id == 3):
        return ((bands_means['B11'] - bands_means['B8A']) + (bands_means['B12'] - bands_means['B4'])) / (bands_means['B11'] + bands_means['B8A'] + bands_means['B12'] + bands_means['B4'])
    if (id == 4):
        return (bands_means['B11'] - bands_means['B4']) / (bands_means['B11'] + bands_means['B4'])


def calculate_normalized_burn_ratio(bands_means, id=None):
    '''
    Calculate the Normalized Burn Ratio (NBR) value for a specific date and polygon in a Sentinel-2 image collection.
    
    The Normalized Burn Ratio is a remote sensing index used to detect and quantify the severity of burn scars
    caused by wildfires. 
    NBR2 is an enhancement of the original Normalized Burn Ratio (NBR) index, which was developed for the same purpose.
    The NBR2 index is sensitive to changes in the vegetation and the charred biomass resulting from a fire, as well as the 
    presence of unburned vegetation and soil background. 
    The values of NBR2 range from -1 to 1, with higher values indicating more severe burn scars. The index is particularly 
    useful in detecting the extent and severity of burn scars in heavily vegetated areas.
    The NBR2 is commonly used in wildfire monitoring and management, as well as in post-fire ecological and land use assessments.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the NBR to be calculated (default is the standard version).
    
    Returns:
        float: the mean NBRx value, for the specified date and polygon.
    '''
    # Calculate NBR
    if (id == None):
        return (bands_means['B8'] - bands_means['B12']) / (bands_means['B8'] + bands_means['B12'])
    elif (id == 2):
        return (bands_means['B11'] - bands_means['B12']) / (bands_means['B11'] + bands_means['B12'])


def calculate_chlorophyll_index(bands_means, id):
    '''
    Calculates the Chlorophyll Index (CI) for a given image, date, and polygon.

    The Chlorophyll Index (CI) is a vegetation index used to estimate chlorophyll content in vegetation. It is calculated
    using the ratio of the near-infrared (NIR) and red (RED) bands of Sentinel-2.
    In general, CI1 is more commonly used and provides a good estimate of chlorophyll content in a variety of vegetation types. 
    However, CI2 may be more suitable for some applications, such as in areas with dense vegetation or where structural properties
    of vegetation are important to consider.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the CI index to be calculated (from 1 to 3).

    Returns:
        float: the calculated CI value, for the specified date and polygon.
    '''
    # Calculate CI (different with respect to the passed id)
    if (id == 1):
        return (bands_means['B8'] / bands_means['B5']) - 1
    elif (id == 2):
        return (bands_means['B8'] / bands_means['B6']) - 1
    elif (id == 3):
        return (bands_means['B8'] / bands_means['B7']) - 1


def calculate_green_coverage_index(bands_means):
    '''
    Calculate the Green Coverage Index (GCI) for a specific date and polygon in a Sentinel-2 image collection.
    
    GCI is a vegetation index that is designed to be less sensitive to atmospheric and soil background effects. It can 
    be used to estimate the fractional green vegetation cover in a given area.
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: the mean value of the GCI, for the specified date and polygon.
    '''
    # Calculate the mean vegetation index for the field polygon
    return (bands_means['B9'] / bands_means['B3']) - 1


def calculate_soil_composition_index(bands_means):
    '''
    Calculates the Soil Composition Index (SCI) for a specified date and polygon in a Sentinel-2 image collection.
    
    The Soil Composition Index (SCI) is a measure of soil composition that is useful for identifying areas with high 
    mineral content, such as desert environments. It is calculated as (B11 - B8) / (B11 + B8).
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: The mean value of the Soil Composition Index (SCI) for the specified date and polygon.
    '''
    # Calculate the mean SCI for the field polygon
    return (bands_means['B11'] - bands_means['B8']) / (bands_means['B11'] + bands_means['B8'])


def calculate_normalized_difference_red_edge(bands_means, id):
    '''
    Calculates the Normalized Difference Red Edge (NDRE) index for a given image, date, and polygon.

    The Normalized Difference Red Edge (NDRE) is a vegetation index used to assess plant chlorophyll and nitrogen
    content. NDRE is sensitive to the presence of healthy, green vegetation and is useful for monitoring crop health,
    stress, and growth.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the NDRE to be calculated (from 1 to 3).

    Returns:
        float: the calculated NDREx value, for the specified date and polygon.
    '''
    # Calculate NDRE
    if (id == 1):
        return (bands_means['B8'] - bands_means['B5']) / (bands_means['B8'] + bands_means['B5'])
    elif (id == 2):
        return (bands_means['B8'] - bands_means['B6']) / (bands_means['B8'] + bands_means['B6'])
    elif (id == 3):
        return (bands_means['B8'] - bands_means['B7']) / (bands_means['B8'] + bands_means['B7'])


def calculate_modified_chlorophyll_absorption_reflectance_index(bands_means, id=None):
    '''
    Calculates the Modified Chlorophyll Absorption in Reflectance Index (MCARI, MCARI1, MCARI2) for a given image, date, and polygon.

    MCARI is a vegetation index used to estimate chlorophyll content in vegetation, and is based on the difference between the 
    green and red reflectances, weighted by the green reflectance. MCARI1 and MCARI2 are modified versions of MCARI, with different 
    weighting factors for the green and red reflectances.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the MCARI index to be calculated (from 0 to 2 - default None).

    Returns:
        float: the calculated MCARI value, for the specified date and polygon.
    '''
    # Calculate MCARI (different with respect to the passed id)
    if (id == None):
        return ((bands_means['B5'] - bands_means['B4']) - 0.2 * (bands_means['B5'] - bands_means['B3'])) * (bands_means['B5'] / bands_means['B4'])
    elif (id == 1):
        return  1.2 * (2.5 * (bands_means['B8'] - bands_means['B4']) - 1.3 * (bands_means['B8'] - bands_means['B3']))
    elif (id == 2):
        return 1.5 * (2.5 * (bands_means['B8'] - bands_means['B4']) - 1.3 * (bands_means['B8'] - bands_means['B3'])) / sqrt(((2.0 * bands_means['B8'] + 1.0) ** 2.0) - (6.0 * bands_means['B8'] - 5.0 * sqrt(bands_means['B4'])) - 0.5)
   

def calculate_chlorophyll_absorption_ratio_index(bands_means, id):
    '''
    Calculates the Chlorophyll Absorption Ratio Index (CARI) for a given image, date, and polygon.

    The Chlorophyll Absorption Ratio Index (CARI) is a vegetation index used to estimate the amount of chlorophyll
    content in vegetation. It is calculated using the ratio of the near-infrared (NIR) and red (RED) bands of Sentinel-2,
    as well as the blue (BLUE) and green (GREEN) bands.

    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
        id (int): The id of the CARI index to be calculated (from 1 to 2).

    Returns:
        float: the calculated CARI value, for the specified date and polygon.
    '''
    # Calculate CARI
    if (id == 1):
        return (bands_means['B5'] / bands_means['B4']) * (sqrt((((bands_means['B5'] - bands_means['B3']) / 150.0 * 670.0 + bands_means['B4'] + (bands_means['B3'] - ((bands_means['B5'] - bands_means['B3']) / 150.0 * 550.0))) ** 2.0))) / (((((bands_means['B5'] - bands_means['B3']) / (150.0 ** 2.0) + 1.0) ** 0.5)))
    elif (id == 2):
        return (abs((((bands_means['B5'] - bands_means['B3']) / 150.0) * bands_means['B4'] + bands_means['B4'] + bands_means['B3'] - (0.496 * bands_means['B3']))) / ((((0.496 ** 2.0) + 1.0) ** 0.5) * (bands_means['B5'] / bands_means['B4'])))


def calculate_bare_soil_index(bands_means):
    '''
    Calculate the Bare Soil Index (BSI) for a specific date and polygon in a Sentinel-2 image collection.
    
    BSI is a vegetation index that can be used to estimate the percentage of bare soil in a given area.
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: the mean value of the BSI, for the specified date and polygon.
    '''
    # Calculate the mean BSI for the field polygon
    return (bands_means['B11'] + bands_means['B4']) + ((bands_means['B8'] + bands_means['B2']) / (bands_means['B11'] + bands_means['B4'])) + (bands_means['B8'] + bands_means['B2'])


def calculate_green_leaf_index(bands_means):
    '''
    Calculate the Green Leaf Index (GLI) for a specific date and polygon in a Sentinel-2 image collection.
    
    GLI is a vegetation index that is designed to be less sensitive to atmospheric and soil background effects. It can 
    be used to estimate the fractional green vegetation cover in a given area.
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: the mean value of the GLI, for the specified date and polygon.
    '''
    # Calculate the mean vegetation index for the field polygon
    return (2.0 * bands_means['B3'] - bands_means['B4'] - bands_means['B2']) / (2.0 * bands_means['B3'] + bands_means['B4'] + bands_means['B2'])


def calculate_alteration_index(bands_means):
    '''
    Calculate the Alteration index for a specific date and polygon in a Sentinel-2 image collection.
    
    Alteration index is used to map the presence of Al-OH or Fe-OH minerals in rocks, and is particularly useful in 
    mapping alteration zones around ore deposits.
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: the mean value of the Alteration index, for the specified date and polygon.
    '''
    # Calculate the mean Alteration index for the field polygon
    return bands_means['B11'] / bands_means['B12']


def calculate_swir_difference_index(bands_means):
    '''
    Calculate the SWIR Difference Index (SDI) for a specific date and polygon in a Sentinel-2 image collection.
    
    SDI is a band ratio index that highlights moisture content and can be used to monitor changes in vegetation
    moisture content, soil moisture, and mineralogical composition.
    
    Args:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    
    Returns:
        float: the mean value of the SDI, for the specified date and polygon.
    '''
    # Calculate the mean SDI for the field polygon
    return bands_means['B8'] - bands_means['B12']


import ee, pandas as pd


def get_all_polarizations(image, date, polygon):
    '''
    Calculate all the mean polarization values for a specific date and polygon in a Sentinel-1 image collection.

    Sentinel-1 is a satellite mission from the European Space Agency (ESA) that carries a Synthetic Aperture Radar (SAR) instrument. 
    The SAR system is capable of transmitting and receiving radar signals in different polarizations, which refers to the orientation 
    of the electromagnetic waves as they travel through space.

    The Sentinel-1 SAR instrument can transmit and receive signals in two different polarization modes:
    - VV (vertical transmit, vertical receive): useful for detecting the roughness of surfaces, such as ocean waves or ice, as the radar 
      signal is more strongly scattered in the vertical direction.
    - VH (vertical transmit, horizontal receive): particularly useful for identifying targets with vertical structures, such as trees 
      and buildings, as it produces strong backscatter signals.

    Args:
        image (ee.ImageCollection): The Sentinel-1 image collection filtered by date and bounds.
        date (pd.Timestamp): The acquisition date in Pandas Timestamp format.
        polygon (ee.Geometry): The field polygon geometry in Earth Engine format.
    
    Returns:
        polarizations_means (dictionary): a dictionary containing for each polarization the mean value, for the specified date and polygon.
    '''
    # Filter image collection to get the image for the date
    image = ee.Image(image.filterDate(date.strftime('%Y-%m-%d'), (date + pd.Timedelta(days=1)).strftime('%Y-%m-%d')).first())

    # Calculate the mean value over the area for the specified polarization type
    polarizations_list = ['VV', 'VH']
    polarizations_means = {}
    for polarization in polarizations_list:
        polarizations_means[polarization] = image.reduceRegion(reducer=ee.Reducer.mean(), geometry=polygon).getInfo()[polarization]
    return polarizations_means


def calculate_radar_vegetation_index(polarizations_means):
    '''
    Calculate the Radar Vegetation Index (RVI) for a specific date and polygon in a Sentinel-1 image collection.
    
    Subhadip Dey: 'Vegetation indices are extremely helpful in understanding the agricultural crop conditions and
    their mapping. In optical remote sensing, Normalized Difference Vegetation Index (NDVI) is one
    such index which conveys such crop health condition within a scene. However, the limitation of
    optical remote sensing is the inability to penetrate cloud cover due to which ground data
    acquisition becomes limited within a crop growth season. Hence, a temporal gap in the acquired
    data might affect crop monitoring as well as crop mapping. On the contrary, Synthetic Aperture
    Radar (SAR) can penetrate the cloud cover, which provides the advantage of acquiring ground data
    during cloudy climatic conditions. Hence, the use of SAR data might be an appealing alternative to
    monitor crop conditions, especially during the monsoon season.'
    Formula source: https://github.com/sentinel-hub/custom-scripts/blob/master/sentinel-1/radar_vegetation_index_code_dual_polarimetric/script.js
    
    Args:
        polarizations_means (dictionary): a dictionary containing for each polarization the mean value, for a specified date and polygon.
    
    Returns:
        float: the mean RVI value, for the specified date and polygon.
    '''
    # Returns RVI
    return (polarizations_means['VH'] * 4.0) / (polarizations_means['VV'] + polarizations_means['VH'])


def calculate_simple_index(polarizations_means, type=''):
    '''
    Calculate a simple index (using VH and VV) for a specific date and polygon in a Sentinel-1 image collection.

    The functions are used to perform arithmetic operations on two bands of the data, specifically the VH and VV bands, 
    which represent the vertical and horizontal polarizations of the radar signal, respectively.
    The functions provide four options for processing the data, depending on the 'type' parameter provided:
    - AVE: This function computes the average of the VH and VV bands, which can be useful for certain applications 
      such as identifying water bodies, where the radar signal is affected by the surface roughness of the water.
    - DIF: This function computes the difference between the VH and VV bands, which can be useful for identifying 
      areas with different types of surfaces, such as vegetation and soil.
    - RAT1: This function computes the ratio of the VV and VH bands, which can be useful for identifying areas 
      with different levels of moisture content, such as wetlands or areas affected by rainfall.
    - RAT2: This function computes the ratio of the VH and VV bands, which can be useful for identifying areas 
      with different levels of roughness, such as urban areas or forests.

    Overall, these functions provide a flexible and powerful set of tools for processing Sentinel-1 radar data, 
    allowing researchers to extract valuable information about the Earth's surface and monitor changes over time.

    Formulae source: https://www.researchgate.net/publication/340548633_Combining_Radar_and_Optical_Imagery_to_Map_Oil_Palm_Plantations_in_Sumatra_Indonesia_Using_the_Google_Earth_Engine

    Args:
        polarizations_means (dictionary): a dictionary containing for each polarization the mean value, for a specified date and polygon.
        type (str): The type of simple operation to get ('AVE', 'DIF', 'RAT1', 'RAT2').

    Returns:
        float: the mean operation value for the specified date and polygon (using VH and VV polarizations).
    '''
    # Return the SIMPLE OPERATION
    if (type == 'AVE'):
        return (polarizations_means['VV'] + polarizations_means['VH']) / 2
    elif (type == 'DIF'):
        return (polarizations_means['VV'] - polarizations_means['VH'])
    elif (type == 'RAT1'):
        return (polarizations_means['VV'] / polarizations_means['VH'])
    elif (type == 'RAT2'):
        return (polarizations_means['VH'] / polarizations_means['VV'])


def calculate_normalized_difference_index(polarizations_means):
    '''
    Calculate the Normalized Difference Index (using VH and VV) for a specific date and polygon in a Sentinel-1 image collection.

    The Normalized Difference Index (NDI) is a commonly used remote sensing index that measures the difference between two spectral bands, 
    which can be helpful in Earth Observation contexts. In the case of Sentinel-1, the NDI can be computed using the VV and VH polarizations, 
    which are sensitive to different types of surfaces and features on the Earth's surface.
    The NDI is important because it can provide valuable information about various features on the Earth's surface, including vegetation, 
    water bodies, and soil moisture. For example, in the case of vegetation, the NDI can be used to estimate the biomass and vegetation health,
    as well as to detect changes in vegetation cover over time. In the case of water bodies, the NDI can be used to estimate the water content 
    and to monitor changes in water levels, which can be useful for flood forecasting and water resource management.
    In the case of soil moisture, the NDI can be used to estimate the soil moisture content, which is important for agriculture and 
    hydrology applications.
    Formula source: https://www.researchgate.net/publication/340548633_Combining_Radar_and_Optical_Imagery_to_Map_Oil_Palm_Plantations_in_Sumatra_Indonesia_Using_the_Google_Earth_Engine

    Args:
        polarizations_means (dictionary): a dictionary containing for each polarization the mean value, for a specified date and polygon.

    Returns:
        float: the mean NDI value for the specified date and polygon.
    '''
    # Calculate NDI
    return (polarizations_means['VV'] - polarizations_means['VH']) / (polarizations_means['VV'] + polarizations_means['VH'])


import ee, pandas as pd


def get_all_bands(image, date, polygon):
    '''
    Calculate all the Landsat-8 bands mean values for a specific date and polygon in a Landsat-8 image collection.

    Spectral bands in Landsat-8 refer to specific ranges of electromagnetic radiation (EMR) that are utilized in remote
    sensing applications to gather information about the Earth's surface. These bands are defined based on their wavelength 
    or frequency and are organized into various groups according to their spectral characteristics, including visible, 
    near-infrared, shortwave infrared, and thermal infrared.
    Each spectral band in Landsat-8 offers distinct insights into the reflectance and absorption properties of features
    on the Earth's surface. For instance, visible bands are particularly sensitive to the reflectance of green vegetation,
    water, and soil. Near-infrared bands are capable of detecting the reflectance of healthy vegetation. Shortwave infrared
    bands enable the detection of variations in moisture content and mineralogy, while thermal infrared bands can capture
    heat signatures.

    Args:
        image (ee.ImageCollection): The Landsat 8 image collection filtered by date and bounds.
        date (pd.Timestamp): The acquisition date in Pandas Timestamp format.
        polygon (ee.Geometry): The field polygon geometry in Earth Engine format.
    
    Returns:
        bands_means (dictionary): a dictionary containing for each band the mean value, for the specified date and polygon.
    '''
    # Filter image collection to get the image for the date
    image = ee.Image(image.filterDate(date.strftime('%Y-%m-%d'), (date + pd.Timedelta(days=1)).strftime('%Y-%m-%d')).first())

    # Returns the list of mean bands values
    bands_list = ['B1', 'B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B8', 'B9', 'B10', 'B11']
    bands_means = {}
    for band in bands_list:
        bands_means[band] = image.reduceRegion(reducer=ee.Reducer.mean(), geometry=polygon).getInfo()[band]
    return bands_means


import ee, pandas as pd, concurrent.futures, sentinel1, sentinel2, landsat8


''' GATHER ALL MEAN INDICES FOR A FIELD, FOR EACH TIME A SATELLITE CONSTELLATION (SENTINEL 1, SENTINEL 2 or LANDSAT 8)
CROSSED THE FIELD, WITHIN A SPECIFIC TIME SPAN '''

def process_chunk(s_filtered, chunk, polygon, field_col_name, field_name, satellite):
    '''
    Processes a chunk of acquisition dates for a given polygon and returns a list of acquisitions.
    
    Parameters:
        s_filtered (ee.ImageCollection): A data object with filtered and pre-processed Sentinel-1, Sentinel-2 or Landsat-8 images.
        chunk (pd.Timestamp list): A list of acquisition dates to process.
        polygon (ee.Geometry): The polygon of interest.
        field_col_name (str): The name of the "field_name" column inside the original DataFrame.
        field_name (str): The name of the field.
        satellite (str): The satellite name ("sentinel-1", "sentinel-2" or "landsat-8").
    
    Returns:
        acquisitions (list): A list of dictionaries, where each dictionary represents an acquisition and contains the 
        acquisition date and associated features.
    '''
    acquisitions = []
    for date in chunk:
        if (satellite == 'sentinel-1'):
            polarizations_means = sentinel1.get_all_polarizations(s_filtered, date, polygon)
            ave = sentinel1.calculate_simple_index(polarizations_means, 'AVE')
            dif = sentinel1.calculate_simple_index(polarizations_means, 'DIF')
            rat1 = sentinel1.calculate_simple_index(polarizations_means, 'RAT1')
            rat2 = sentinel1.calculate_simple_index(polarizations_means, 'RAT2')
            ndi = sentinel1.calculate_normalized_difference_index(polarizations_means)
            rvi = sentinel1.calculate_radar_vegetation_index(polarizations_means)

            # Create a dataframe row for the date
            df_acquisition= {str(field_col_name): field_name, 's1_acquisition_date': date,
                        'VV': polarizations_means['VV'], 'VH': polarizations_means['VH'], 
                        'AVE': ave, 'DIF': dif, 'RAT1': rat1, 'RAT2': rat2, 
                        'NDI': ndi, 'RVI': rvi}
        
        elif (satellite == 'sentinel-2'):
            # Calculate standard bands
            bands_means = sentinel2.get_all_bands(s_filtered, date, polygon)
            # Calculate vegetation indexes
            ndvi = sentinel2.calculate_vegetation_index(bands_means, 'ND')
            nsndvi = sentinel2.calculate_vegetation_index(bands_means, 'NSND')
            gndvi = sentinel2.calculate_vegetation_index(bands_means, 'GND')
            rendvi1 = sentinel2.calculate_vegetation_index(bands_means, 'REND', 1)
            rendvi2 = sentinel2.calculate_vegetation_index(bands_means, 'REND', 2)
            rendvi3 = sentinel2.calculate_vegetation_index(bands_means, 'REND', 3)
            grndvi = sentinel2.calculate_vegetation_index(bands_means, 'GRND')
            gbndvi = sentinel2.calculate_vegetation_index(bands_means, 'GBND')
            savi = sentinel2.calculate_vegetation_index(bands_means, 'SA')
            osavi = sentinel2.calculate_vegetation_index(bands_means, 'OSA')
            msavi = sentinel2.calculate_vegetation_index(bands_means, 'MSA')
            tsavi = sentinel2.calculate_vegetation_index(bands_means, 'TSA')
            atsavi = sentinel2.calculate_vegetation_index(bands_means, 'ATSA')
            rvi = sentinel2.calculate_vegetation_index(bands_means, 'R')
            dvi = sentinel2.calculate_vegetation_index(bands_means, 'D')
            cvi = sentinel2.calculate_vegetation_index(bands_means, 'C')
            ctvi = sentinel2.calculate_vegetation_index(bands_means, 'CT')
            avi = sentinel2.calculate_vegetation_index(bands_means, 'A')
            arvi1 = sentinel2.calculate_vegetation_index(bands_means, 'AR', 1)
            arvi2 = sentinel2.calculate_vegetation_index(bands_means, 'AR', 2)
            evi1 = sentinel2.calculate_vegetation_index(bands_means, 'E', 1)
            evi2 = sentinel2.calculate_vegetation_index(bands_means, 'E', 2)
            evi3 = sentinel2.calculate_vegetation_index(bands_means, 'E', 3)
            wdrvi = sentinel2.calculate_vegetation_index(bands_means, 'WDR')
            mtvi1 = sentinel2.calculate_vegetation_index(bands_means, 'MT', 1)
            mtvi2 = sentinel2.calculate_vegetation_index(bands_means, 'MT', 2)
            # Calculate exogenous organic matter indexes
            eomi1 = sentinel2.calculate_exogenous_organic_matter_index(bands_means, 1)
            eomi2 = sentinel2.calculate_exogenous_organic_matter_index(bands_means, 2)
            eomi3 = sentinel2.calculate_exogenous_organic_matter_index(bands_means, 3)
            eomi4 = sentinel2.calculate_exogenous_organic_matter_index(bands_means, 4)
            # Calculate normalized burn ratio indexes
            nbr = sentinel2.calculate_normalized_burn_ratio(bands_means)
            nbr2 = sentinel2.calculate_normalized_burn_ratio(bands_means, 2)
            # Calculate clorophyl indexes
            ci1 = sentinel2.calculate_chlorophyll_index(bands_means, 1)
            ci2 = sentinel2.calculate_chlorophyll_index(bands_means, 2)
            ci3 = sentinel2.calculate_chlorophyll_index(bands_means, 3)
            # Calculate green coverage index
            gci = sentinel2.calculate_green_coverage_index(bands_means)
            # Calculate soil composition index
            sci = sentinel2.calculate_soil_composition_index(bands_means)
            # Calculate normalized difference red edge
            ndre1 = sentinel2.calculate_normalized_difference_red_edge(bands_means, 1)
            ndre2 = sentinel2.calculate_normalized_difference_red_edge(bands_means, 2)
            ndre3 = sentinel2.calculate_normalized_difference_red_edge(bands_means, 3)
            # Calculate chlorophyll absorption ratio index
            cari1 = sentinel2.calculate_chlorophyll_absorption_ratio_index(bands_means, 1)
            cari2 = sentinel2.calculate_chlorophyll_absorption_ratio_index(bands_means, 2)
            # Calculate modified chlorophyll absorption in reflectance index
            mcari = sentinel2.calculate_modified_chlorophyll_absorption_reflectance_index(bands_means)
            mcari1 = sentinel2.calculate_modified_chlorophyll_absorption_reflectance_index(bands_means, 1)
            mcari2 = sentinel2.calculate_modified_chlorophyll_absorption_reflectance_index(bands_means, 2)
            # Calculate bare soil index
            bsi = sentinel2.calculate_bare_soil_index(bands_means)
            # Calculate green leaf index
            gli = sentinel2.calculate_green_leaf_index(bands_means)
            # Compute alteration index
            alteration = sentinel2.calculate_alteration_index(bands_means)
            # Compute SWIR Difference Index
            sdi = sentinel2.calculate_swir_difference_index(bands_means)

            # Create a dataframe row for the date
            df_acquisition= {str(field_col_name): field_name, 's2_acquisition_date': date, 
                        'B1': bands_means['B1'], 'B2': bands_means['B2'], 'B3': bands_means['B3'], 'B4': bands_means['B4'], 'B5': bands_means['B5'],
                        'B6': bands_means['B6'], 'B7': bands_means['B7'], 'B8': bands_means['B8'], 'B8A': bands_means['B8A'], 'B9': bands_means['B9'],
                        'B11': bands_means['B11'], 'B12': bands_means['B12'],     
                        'NDVI': ndvi, 'NSNDVI': nsndvi, 'GNDVI': gndvi, 'RENDVI1': rendvi1, 'RENDVI2': rendvi2,
                        'RENDVI3': rendvi3, 'GRNDVI': grndvi, 'GBNDVI': gbndvi,
                        'SAVI': savi, 'OSAVI': osavi, 'MSAVI': msavi, 'TSAVI': tsavi, 'ATSAVI': atsavi,
                        'RVI': rvi, 'DVI': dvi, 'CVI': cvi, 'CTVI': ctvi, 'AVI': avi, 'ARVI1': arvi1, 'ARVI2': arvi2,
                        'EVI1': evi1, 'EVI2': evi2, 'EVI3': evi3, 'WDRVI': wdrvi, 'MTVI1': mtvi1, 'MTVI2': mtvi2,
                        'EOMI1': eomi1, 'EOMI2': eomi2, 'EOMI3': eomi3, 'EOMI4': eomi4, 
                        'NBR': nbr, 'NBR2': nbr2,
                        'CI1': ci1, 'CI2': ci2, 'CI3': ci3,
                        'GCI': gci, 'SCI': sci, 'NDRE1': ndre1, 'NDRE2': ndre2, 'NDRE3': ndre3,
                        'CARI1': cari1, 'CARI2': cari2, 'MCARI': mcari, 'MCARI1': mcari1, 'MCARI2': mcari2,
                        'BSI': bsi, 'GLI': gli, 'ALTERATION': alteration, 'SDI': sdi}
        
        if (satellite == 'landsat-8'):
            bands_means = landsat8.get_all_bands(s_filtered, date, polygon)

            # Create a dataframe row for the date
            df_acquisition= {str(field_col_name): field_name, 'l8_acquisition_date': date,
                        'B1': bands_means['B1'], 'B2': bands_means['B2'], 'B3': bands_means['B3'],
                        'B4': bands_means['B4'], 'B5': bands_means['B5'], 'B6': bands_means['B6'],
                        'B7': bands_means['B7'], 'B8': bands_means['B8'], 'B9': bands_means['B9'],
                        'B10': bands_means['B10'], 'B11': bands_means['B11']}
            
        # Add row to the list
        acquisitions.append(df_acquisition)

    return acquisitions


def parallelize_features_read(s_filtered, date_range, polygon, field_col_name, field_name, satellite, already_occupied_threads):
    '''
    Reads Sentinel-1, Sentinel-2 or Landsat-8 images and gathers spectral features.
    Furthermore, it makes multiple I/O requests to GEE in parallel - otherwise to compute the same task would have taken too much
    time. Each thread works on different date chunks, maintaining fixed the field.

    Parameters:
        s_filtered (ee.ImageCollection): A data object with filtered and pre-processed Sentinel-1, Sentinel-2 or Landsat-8 images.
        date_range (pd.Timestamp list): A list of acquisition dates to process.
        polygon (ee.Geometry): The polygon of interest.
        field_col_name (str): The name of the "field_name" column inside the original DataFrame.
        field_name (str): The name of the field.
        satellite (str): The satellite name ("sentinel-1", "sentinel-2" or "landsat-8").
        already_occupied_threads (int): The number of threads dedicated to parallelization over the field level.
        
    Returns:
        acquisitions (list): A list of dictionaries, where each dictionary represents an acquisition and contains the 
        acquisition date and associated features.
    '''
    # Split date range into multiple chunks
    ThreadPoolExecutor = concurrent.futures.ThreadPoolExecutor()
    num_chunks = ThreadPoolExecutor._max_workers - already_occupied_threads
    chunk_size = len(date_range) // num_chunks
    date_chunks = [date_range[i:i+chunk_size] for i in range(0, len(date_range), chunk_size)]

    # Create threads to process date chunks in parallel
    acquisitions = []
    # Each thread works on a different chunks of dates to gather features
    with concurrent.futures.ThreadPoolExecutor(num_chunks) as executor:
        futures = []
        for chunk in date_chunks:
            futures.append(executor.submit(process_chunk, s_filtered, chunk, polygon, field_col_name, field_name, satellite))
                
        # Wait for all futures to complete and gather results
        for future in concurrent.futures.as_completed(futures):
            acquisitions.extend(future.result())

    return acquisitions


def process_field(field, start_date, end_date, satellite, filters_params, already_occupied_threads):
    '''
    Processes a single field by calculating the mean indices for each time the satellite passed over that field.

    Args:
        field (pandas Dataframe - row): The field item containing the field name and polygon coordinates.
        start_date (str): The start date of the date range to filter the collection by.
        end_date (str): The end date of the date range to filter the collection by.
        satellite (str): The satellite name ("sentinel-1", "sentinel-2" or "landsat-8").
        filters_params (list): The list of parameters to be used for filters parameters for extracting satellite Images Collections.
        already_occupied_threads (int): The number of threads dedicated to parallelization over the field level.

    Returns:
        list: A list of dictionary objects containing the mean calculated indices for each acquisition date, for the
        specified field (using the selected satellite).
    '''
    # Get the field name and polygon coordinates
    field_name = field[0]
    polygon = ee.Geometry.Polygon(field[1])
    
    if (satellite == 'sentinel-1'):
        # Filter Sentinel 1 collection
        s_collection = ee.ImageCollection('COPERNICUS/S1_GRD')
        s_filtered = s_collection.filterBounds(polygon).filterDate(str(start_date), str(end_date)) \
                                    .filter(ee.Filter.listContains('transmitterReceiverPolarisation', 'VV')) \
                                    .filter(ee.Filter.listContains('transmitterReceiverPolarisation', 'VH')) \
                                    .filter(ee.Filter.eq('instrumentMode', 'IW')) \
                                    .filter(ee.Filter.eq('orbitProperties_pass', filters_params[0])) #\
                                    #.filter(ee.Filter.eq('relativeOrbitNumber_start', 66)) \
                                    #.filterMetadata('resolution_meters', 'equals', 10)
    elif (satellite == 'sentinel-2'):
        # Filter Sentinel 2 collection
        s_collection = ee.ImageCollection('COPERNICUS/S2_SR')
        s_filtered = s_collection.filterBounds(polygon).filterDate(str(start_date), str(end_date)) \
                                .filter(ee.Filter.lte('CLOUDY_PIXEL_PERCENTAGE', int(filters_params[0])))
    elif (satellite == 'landsat-8'):
        # Filter Landsat 8 collection
        s_collection = ee.ImageCollection('LANDSAT/LC08/C02/T1_TOA')
        s_filtered = s_collection.filterBounds(polygon).filterDate(str(start_date), str(end_date))
        
    # Get distinct dates from the satellite Image collection and put into the date_range list
    s_dates = s_filtered.aggregate_array('system:time_start').map(lambda time_start: ee.Date(time_start).format('YYYY-MM-dd'))
    s_dates_distinct = s_dates.distinct().sort()
    date_range = pd.to_datetime(s_dates_distinct.getInfo())
    
    return parallelize_features_read(s_filtered, date_range, polygon, field.index[0], field_name, satellite, already_occupied_threads)


def get_features(fields_df, start_date, end_date, satellite, filters_params=None, fields_threads=4):
    '''
    It allows to get from a pandas DataFrame composed of crop fields information, another DataFrame that contains 
    for each time a satellite (sentinel-1, sentinel-2 or landsat-8) passed on regions of interest, within a given time period, 
    all the mainly used features mean values (optical, radar or thermal).
    Furthermore, it makes multiple I/O requests to Google Earth Engine (GEE) in parallel - otherwise to compute the same task 
    would have taken too much time. Each thread works on different date chunks, maintaining fixed the field (obviously if you
    have only one field this part is not parallelized).
    Please consider that parallelization of requests has been applied on two levels: on fields and on dates.

    Args:
        fields_df (pandas DataFrame): A DataFrame containing the crop field name and polygon coordinates for each field.
        start_date (str): The start date of the date range to filter the collection by.
        end_date (str): The end date of the date range to filter the collection by.
        satellite (str): The satellite name ("sentinel-1": radar, "sentinel-2": optical, "landsat-8": optical + thermal).
        filters_params (list - default `None`): The list of parameters values to be used for filters to extracting satellite Images Collections.
            * For Sentinel-1:
                * first parameter in the list rapresents the value of the 'orbitProperties_pass' filter ('ASCENDING' or 'DESCENDING')
            * For Sentinel-2:
                * first parameter in the list represents the value of the 'CLOUDY_PIXEL_PERCENTAGE' filter ('LTE' - values in range [0, 100])
        fields_threads (int - default `4`): The number of threads to dedicate to parallelization of requests over the fields level, the remaining part 
            instead is used to apply parallelization over dates level. The value of this parameter should be high (with respect 
            to the overall number of threads exploitable) if you have a lot of crop fields but a little time-span to consider, whereas
            if you have fewer fields but a bigger time-span you should decrease this parameter.
            Finally, if you have lot of fields with lot of dates to process it should may be optimal considering half of the 
            overall number of threads available.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the calculated indices for each acquisition date, for each field within
        the specified date range and using the selected satellite. The DataFrame includes columns for the crop 
        details, acquisition date and calculated indices. The calculated indices depend on the selected satellite.
    '''
    # Create an empty list to store the data for each field
    df_list = []

    # Calculate all the indices for each field, for the selected time period
    # In parallell, to improve performances (each thread works on one single field)
    with concurrent.futures.ThreadPoolExecutor(fields_threads) as executor:
        futures = []
        # When just one field is inside the fields_df DataFrame
        if (len(fields_df) == 2):
            futures.append(executor.submit(process_field, fields_df, start_date, end_date, satellite, filters_params, fields_threads))
        else:
            for index, field in fields_df.iterrows():
                futures.append(executor.submit(process_field, field, start_date, end_date, satellite, filters_params, fields_threads))
                
        # Wait for all futures to complete and gather results
        for future in concurrent.futures.as_completed(futures):
            df_list.extend(future.result())

    # Create a dataframe from the list of rows
    df = pd.DataFrame(df_list).reset_index(drop=True)
    
    # Sorting by crop field name and acquisition date and return the sorted DataFrame
    return df.sort_values([str(df.columns[0]), str(df.columns[1])], ascending=[True, True])